﻿/**
 * Copyright (c) 2012, Lokkju <lokkju@lokkju.com>
 * Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted,
 * provided that the above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * A console tool for appending WGS84 lat/lon coordinates to each line in a csv file 
 * containing SR-ORG:6856 x/y values.
 * 
 * SR-ORG:6856 is NAD 1983 HARN StatePlane Oregon North FIPS 3601 Feet Intl.

 * Usage: srorg6856_tool [options] -i <input csv>
 * Options:
 *  -i, --input=CSV_FILE       the input CSV_FILE to process (REQUIRED).
 *  -o, --output=FILE          the output FILE to write to.
 *  -x, --xfield=X             the field name or index of the X value.
 *  -y, --yfield=Y             the field name or index of the Y value.
 *  -v                         increase debug message verbosity
 *  -h, -?, --help             print this help and exit
 *  
 */

using System;
using System.Collections.Generic;
using System.Text;
using DotSpatial.Projections;
using Mono.Options;
using System.IO;
namespace com.lokkju.sharp.geotools.srorg6856_tool
{
    class Program
    {
        static int verbosity = 0;
        static void Main(string[] args)
        {
            // ProjectionInfo epsg2913 = ProjectionInfo.FromEsriString("PROJCS[\"NAD_1983_HARN_StatePlane_Oregon_North_FIPS_3601_Feet_Intl\", GEOGCS[\"GCS_North_American_1983_HARN\", DATUM[\"D_North_American_1983_HARN\", SPHEROID[\"GRS_1980\", 6378137.0, 298.257222101]], PRIMEM[\"Greenwich\", 0.0], UNIT[\"Degree\", 0.017453292519943295]], PROJECTION[\"Lambert_Conformal_Conic\"], PARAMETER[\"False_Easting\", 8202099.737532808], PARAMETER[\"False_Northing\", 0.0], PARAMETER[\"Central_Meridian\", -120.5], PARAMETER[\"Standard_Parallel_1\", 44.33333333333334], PARAMETER[\"Standard_Parallel_2\", 46.0], PARAMETER[\"Latitude_Of_Origin\", 43.66666666666666], UNIT[\"Foot\", 0.3048]]");            
            ProjectionInfo srorg6856 = KnownCoordinateSystems.Projected.StatePlaneNad1983HarnFeet.NAD1983HARNStatePlaneOregonNorthFIPS3601FeetIntl;
            ProjectionInfo wgs1984 = KnownCoordinateSystems.Geographic.World.WGS1984;
            ProjectionInfo webMercator = KnownCoordinateSystems.Projected.World.WebMercator;

            bool show_help = false;
            string o_field_x = "x";
            string o_field_y = "y";
            int o_field_x_pos = 0;
            int o_field_y_pos = 0;
            string o_input_file = "";
            string o_output_file = "";
            var p = new OptionSet(){
                { "i=|input=", "the input {CSV_FILE} to process (REQUIRED).", v => o_input_file = v },
                { "o=|output=","the output {FILE} to write to.", v => o_output_file = v },
                { "x=|xfield=","the field name or index of the {X} value.", v => o_field_x = v  },
                { "y=|yfield=","the field name or index of the {Y} value.", v => o_field_y = v  },
                { "v", "increase debug message verbosity",v => { if (v != null) ++verbosity; } },
                { "h|?|help","print this help and exit", v => show_help = v != null  },
            };
            List<string> extra;
            try {
                extra = p.Parse(args);
            } catch (OptionException e) {
                Error("An Exception has occured: {0}", e.Message);
                Error("Try 'srorg6856_tool --help' for more information.");
                return;
            }

            if (show_help)
            {
                ShowHelp(p);
                return;
            }
            // require input csv
            if (String.IsNullOrEmpty(o_input_file) || !File.Exists(o_input_file))
            {
                Error("Error: an input CSV file is required");
                Error("Try 'srorg6856_tool --help' for more information.");
                return;
            }
            Debug("Using input file: {0}", o_input_file);
            // find field indexes
            if (Int32.TryParse(o_field_x,out o_field_x_pos) && Int32.TryParse(o_field_y,out o_field_y_pos)) {
                Debug("Using field indexes x:{0} and y:{1}",o_field_x_pos,o_field_y_pos);
            } else {
                Debug("Finding field indexes for fields '{0}' and '{1}'",o_field_x,o_field_y);
                if (!findHeaderIndexes(o_input_file, o_field_x, o_field_y, out o_field_x_pos, out o_field_y_pos))
                {
                    Error("Error: could not find field indexes for the specified X and Y fields.");
                    Error("Try 'srorg6856_tool --help' for more information.");
                    return;
                }
                Debug("Using field indexes x:{0} and y:{1}", o_field_x_pos, o_field_y_pos);
            }
            TextWriter output = Console.Out;
            if (!String.IsNullOrWhiteSpace(o_output_file) && !File.Exists(o_output_file))
            {

                try
                {
                    output = File.CreateText(o_output_file);
                    Debug("Writing to '{0}'", o_output_file);
                }
                catch (Exception e)
                {
                    Error("Error: Could not open output file '{0}' for writing: {1}", o_output_file,e.Message);
                    Error("Try 'srorg6856_tool --help' for more information.");
                    return;
                }
            }
            else if (File.Exists(o_output_file))
            {
                Error("Output file '{0}' already exists, not overwriting.", o_output_file);
                Error("Try 'srorg6856_tool --help' for more information.");
                return;
            }
            else
            {
                Debug("Writing output to Console stdout");
            }
            
            using (StreamReader reader = new StreamReader(o_input_file))
            {
                // process the header line
                string line = reader.ReadLine();
                output.WriteLine("{0},\"lat\",\"lon\"", line);

                // process the rest of the lines
                int ln = 0;
                while (true)
                {
                    ln++;
                    line = reader.ReadLine();
                    if (line == null) break;
                    try
                    {
                        string[] parts = line.Split(new char[] { ',' });
                        double lx = Double.Parse(parts[o_field_x_pos].Trim(new char[] { '"', ' ' }));
                        double ly = Double.Parse(parts[o_field_y_pos].Trim(new char[] { '"', ' ' }));
                        double[] dd_xy = new double[] { lx, ly };
                        double[] wm_xy = new double[] { lx, ly };
                        // TODO: may need to use second form of this method?
                        Reproject.ReprojectPoints(dd_xy, new double[1], srorg6856, wgs1984, 0, 1);
                        //Reproject.ReprojectPoints(wm_xy, new double[1], srorg6856, webMercator, 0, 1);
                        //Console.WriteLine("{0},\"{1}\",\"{2}\",\"{3}\",\"{4}\"", line, dd_xy[0], dd_xy[1], wm_xy[0], wm_xy[1]);
                        output.WriteLine("{0},\"{1}\",\"{2}\"", line, dd_xy[0], dd_xy[1]);
                    } 
                    catch (Exception e)
                    {
                        output.WriteLine("{0},\"\",\"\",\"\",\"\"", line);
                        Error("Line #{0} is improprly formatted, skipping conversion: {1}", ln, e.Message);
                    }
                }
            }

        }
        static void ShowHelp(OptionSet p)
        {
            Console.WriteLine("srorg6856_tool (C)2012 Lokkju <loki@lokkju.com>");
            Console.WriteLine("Released under the ISC License.");
            Console.WriteLine("Using the Dotspatial.Projections library under LGPL.");
            Console.WriteLine();
            Console.WriteLine("Append Lat/Lon Decimal Degree coordinates to each line");
            Console.WriteLine("in a CSV file, based on the SR-ORG:6856 X/Y values on that line.");
            Console.WriteLine("SR-ORG:6856 is NAD 1983 HARN StatePlane Oregon North FIPS 3601 Feet Intl.");
            Console.WriteLine();
            Console.WriteLine("Usage: srorg6856_tool [options] -i <input csv>");
            Console.WriteLine();
            Console.WriteLine("Options:");
            p.WriteOptionDescriptions(Console.Out);
        }
        static void Debug(string format, params object[] args)
        {
            if (verbosity > 0)
            {
                TextWriter errorWriter = Console.Error;
                errorWriter.Write("# ");
                errorWriter.WriteLine(format, args);
            }
        }
        static void Error(string format, params object[] args)
        {
            TextWriter errorWriter = Console.Error;
            errorWriter.Write("# ");
            errorWriter.WriteLine(format, args);
        }

        static bool findHeaderIndexes(string csv_file,string x,string y,out int xi,out int yi) {
            xi = 0;
            yi = 0;
            bool x_found = false;
            bool y_found = false;
	        using (StreamReader reader = new StreamReader(csv_file))
	        {
	            string line = reader.ReadLine();
		        if (line == null) throw new Exception("No lines found in CSV file");
        		string[] parts = line.Split(new char[]{','});
                for (int i=0;i<parts.Length;i++) {
                    string t = parts[i].Trim(new char[]{'"',' '});
                    if ( String.Equals(t,x,StringComparison.OrdinalIgnoreCase)) {
                        if(!x_found) {
                            x_found = true;
                            xi = i;
                        } else {
                            Debug("Two fields with the same name were found for X");
                            return false;
                        }
                    }
                    if ( String.Equals(t,y,StringComparison.OrdinalIgnoreCase)) {
                        if(!y_found) {
                            y_found = true;
                            yi = i;
                        } else {
                            Debug("Two fields with the same name were found for Y");
                            return false;
                        }
                    }
                }
    	    }
            if (!x_found) {
                Debug("The field index for X field '{0}' could not be found",x);
                return false;
            }
            if (!y_found) {
                Debug("The field index for Y field '{0}' could not be found",y);
                return false;
            }
            if (xi == yi)
            {
                Debug("The field indexes for X and Y are the same.  This isn't really possible...");
                return false;
            }
            return true;
        }
    }
}
